@@ -277,6 +277,9 @@ def launch(boss: Boss, opts: LaunchCLIOptions, args: List[str], target_tab: Opti
         kw['overlay_for'] = active.id
     if opts.stdin_source and opts.stdin_source != 'none':
         q = opts.stdin_source
+        if opts.stdin_add_formatting:
+            if q in ('@screen', '@screen_scrollback', '@alternate', '@alternate_scrollback'):
+                q = '@ansi_' + q[1:]
         if opts.stdin_add_line_wrap_markers:
             q += '_wrap'
         penv, stdin = boss.process_stdin_source(window=active, stdin=q)
