@@ -15,6 +15,8 @@
 #include "control-codes.h"
 #include "monotonic.h"
 
+extern PyTypeObject Screen_Type;
+
 static MouseShape mouse_cursor_shape = BEAM;
 typedef enum MouseActions { PRESS, RELEASE, DRAG, MOVE } MouseAction;
 
@@ -95,37 +97,37 @@ encode_mouse_event(Window *w, int button, MouseAction action, int mods) {
 
 // }}}
 
-static inline double
-window_left(Window *w, OSWindow *os_window) {
-    return w->geometry.left - OPT(window_padding_width) * (os_window->logical_dpi_x / 72.0);
+static inline unsigned int
+window_left(Window *w) {
+    return w->geometry.left - w->padding.left;
 }
 
-static inline double
-window_right(Window *w, OSWindow *os_window) {
-    return w->geometry.right + OPT(window_padding_width) * (os_window->logical_dpi_x / 72.0);
+static inline unsigned int
+window_right(Window *w) {
+    return w->geometry.right + w->padding.right;
 }
 
-static inline double
-window_top(Window *w, OSWindow *os_window) {
-    return w->geometry.top - OPT(window_padding_width) * (os_window->logical_dpi_y / 72.0);
+static inline unsigned int
+window_top(Window *w) {
+    return w->geometry.top - w->padding.top;
 }
 
-static inline double
-window_bottom(Window *w, OSWindow *os_window) {
-    return w->geometry.bottom + OPT(window_padding_width) * (os_window->logical_dpi_y / 72.0);
+static inline unsigned int
+window_bottom(Window *w) {
+    return w->geometry.bottom + w->padding.bottom;
 }
 
 static inline bool
-contains_mouse(Window *w, OSWindow *os_window) {
+contains_mouse(Window *w) {
     double x = global_state.callback_os_window->mouse_x, y = global_state.callback_os_window->mouse_y;
-    return (w->visible && window_left(w, os_window) <= x && x <= window_right(w, os_window) && window_top(w, os_window) <= y && y <= window_bottom(w, os_window));
+    return (w->visible && window_left(w) <= x && x <= window_right(w) && window_top(w) <= y && y <= window_bottom(w));
 }
 
 static inline double
-distance_to_window(Window *w, OSWindow *os_window) {
+distance_to_window(Window *w) {
     double x = global_state.callback_os_window->mouse_x, y = global_state.callback_os_window->mouse_y;
-    double cx = (window_left(w, os_window) + window_right(w, os_window)) / 2.0;
-    double cy = (window_top(w, os_window) + window_bottom(w, os_window)) / 2.0;
+    double cx = (window_left(w) + window_right(w)) / 2.0;
+    double cy = (window_top(w) + window_bottom(w)) / 2.0;
     return (x - cx) * (x - cx) + (y - cy) * (y - cy);
 }
 
@@ -140,7 +142,7 @@ cell_for_pos(Window *w, unsigned int *x, unsigned int *y, bool *in_left_half_of_
     bool in_left_half = true;
     double mouse_x = global_state.callback_os_window->mouse_x;
     double mouse_y = global_state.callback_os_window->mouse_y;
-    double left = window_left(w, os_window), top = window_top(w, os_window), right = window_right(w, os_window), bottom = window_bottom(w, os_window);
+    double left = window_left(w), top = window_top(w), right = window_right(w), bottom = window_bottom(w);
     if (clamp_to_window) {
         mouse_x = MIN(MAX(mouse_x, left), right);
         mouse_y = MIN(MAX(mouse_y, top), bottom);
@@ -511,7 +513,7 @@ window_for_event(unsigned int *window_idx, bool *in_tab_bar) {
     if (!*in_tab_bar && global_state.callback_os_window->num_tabs > 0) {
         Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
         for (unsigned int i = 0; i < t->num_windows; i++) {
-            if (contains_mouse(t->windows + i, global_state.callback_os_window) && t->windows[i].render_data.screen) {
+            if (contains_mouse(t->windows + i) && t->windows[i].render_data.screen) {
                 *window_idx = i; return t->windows + i;
             }
         }
@@ -527,7 +529,7 @@ closest_window_for_event(unsigned int *window_idx) {
         Tab *t = global_state.callback_os_window->tabs + global_state.callback_os_window->active_tab;
         for (unsigned int i = 0; i < t->num_windows; i++) {
             Window *w = t->windows + i;
-            double d = distance_to_window(w, global_state.callback_os_window);
+            double d = distance_to_window(w);
             if (d < closest_distance) { ans = w; closest_distance = d; *window_idx = i; }
         }
     }
@@ -698,6 +700,25 @@ scroll_event(double UNUSED xoffset, double yoffset, int flags) {
     }
 }
 
+static PyObject*
+send_mouse_event(PyObject *self UNUSED, PyObject *args) {
+    Screen *screen;
+    unsigned int x, y;
+    int button, action, mods;
+    if (!PyArg_ParseTuple(args, "O!IIiii", &Screen_Type, &screen, &x, &y, &button, &action, &mods)) return NULL;
+
+    MouseTrackingMode mode = screen->modes.mouse_tracking_mode;
+    if (mode == ANY_MODE || (mode == MOTION_MODE && action != MOVE) || (mode == BUTTON_MODE && (action == PRESS || action == RELEASE))) {
+        int sz = encode_mouse_event_impl(x + 1, y + 1, screen->modes.mouse_tracking_protocol, button, action, mods);
+        if (sz > 0) {
+            mouse_event_buf[sz] = 0;
+            write_escape_code_to_child(screen, CSI, mouse_event_buf);
+            Py_RETURN_TRUE;
+        }
+    }
+    Py_RETURN_FALSE;
+}
+
 static PyObject*
 test_encode_mouse(PyObject *self UNUSED, PyObject *args) {
     unsigned int x, y;
@@ -732,6 +753,7 @@ send_mock_mouse_event_to_window(PyObject *self UNUSED, PyObject *args) {
 }
 
 static PyMethodDef module_methods[] = {
+    METHODB(send_mouse_event, METH_VARARGS),
     METHODB(test_encode_mouse, METH_VARARGS),
     METHODB(send_mock_mouse_event_to_window, METH_VARARGS),
     {NULL, NULL, 0, NULL}        /* Sentinel */
@@ -739,6 +761,10 @@ static PyMethodDef module_methods[] = {
 
 bool
 init_mouse(PyObject *module) {
+    PyModule_AddIntMacro(module, PRESS);
+    PyModule_AddIntMacro(module, RELEASE);
+    PyModule_AddIntMacro(module, DRAG);
+    PyModule_AddIntMacro(module, MOVE);
     if (PyModule_AddFunctions(module, module_methods) != 0) return false;
     return true;
 }
